from flask import Flask, request
import csv
from datetime import datetime
import logging
import os
from collections import Counter
from flask import render_template_string, send_file
import io
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import json

app = Flask(__name__)
# Add detailed logging for Flask errors and requests
# Also log to console for visibility in systemd logs
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
console.setFormatter(formatter)
logging.getLogger().addHandler(console)
@app.before_request
def before_request_logging():
    logging.debug(f"Handling request: {request.method} {request.url} Headers: {dict(request.headers)}")

@app.after_request
def after_request_logging(response):
    logging.debug(f"Response status: {response.status} for {request.method} {request.url}")
    return response

@app.teardown_request
def teardown_request_logging(exception):
    if exception:
        logging.error(f"Teardown exception: {exception}")
logging.basicConfig(filename='/srv/http/Requests/server.log', level=logging.DEBUG, 
                    format='%(asctime)s %(levelname)s: %(message)s')

@app.route('/api/log', methods=['POST'])
def log_addon():
    logging.debug(f"Received request: {request.method} {request.url} Content-Type: {request.content_type}")
    try:
        # Check Content-Type
        if request.content_type != 'application/json':
            logging.error(f"Invalid Content-Type: {request.content_type}")
            return {'status': 'error', 'message': 'Content-Type must be application/json'}, 400
        
        # Safely parse JSON
        try:
            data = request.get_json(force=True)
            logging.debug(f"Parsed JSON data: {data}")
        except Exception as e:
            logging.error(f"Failed to parse JSON: {str(e)}")
            return {'status': 'error', 'message': 'Invalid JSON payload'}, 400

        addon = data.get('addon')
        user_uuid = data.get('uuid')
        link = data.get('link')  # New: get link if present
        if not addon or not isinstance(addon, str):
            logging.error(f"Invalid addon name: {addon}")
            return {'status': 'error', 'message': 'Invalid or missing addon name'}, 400
        if not user_uuid or not isinstance(user_uuid, str):
            logging.error(f"Invalid uuid: {user_uuid}")
            return {'status': 'error', 'message': 'Invalid or missing uuid'}, 400

        from datetime import timezone
        timestamp = datetime.now(timezone.utc).isoformat()
        log_file = '/srv/http/Requests/addon_log.csv'
        os.makedirs(os.path.dirname(log_file), exist_ok=True)
        with open(log_file, 'a', newline='') as f:
            writer = csv.writer(f)
            writer.writerow([addon, user_uuid, timestamp, link if link else ""])
        logging.info(f"Logged addon: {addon} uuid: {user_uuid} at {timestamp} link: {link}")
        return {'status': 'logged'}, 200

    except Exception as e:
        logging.error(f"Unexpected error: {str(e)}")
        return {'status': 'error', 'message': str(e)}, 500

def empty_chart_message(msg="No data available"):
    plt.figure(figsize=(6,2))
    plt.text(0.5, 0.5, msg, ha='center', va='center', fontsize=14)
    plt.axis('off')
    img = io.BytesIO()
    plt.savefig(img, format='png')
    plt.close()
    img.seek(0)
    return send_file(img, mimetype='image/png')

# Load UUID blacklist once at startup
UUID_BLACKLIST = set()
try:
    with open('uuid_blacklist.json', 'r', encoding='utf-8') as f:
        UUID_BLACKLIST = set(json.load(f))
except Exception:
    UUID_BLACKLIST = set()

def is_uuid_blacklisted(uuid):
    return uuid in UUID_BLACKLIST

def parse_entries(log_file):
    entries = []
    if os.path.exists(log_file):
        with open(log_file, newline='') as f:
            reader = csv.reader(f)
            for row in reader:
                # Support both old (3 fields) and new (4 fields) rows
                if len(row) == 4:
                    entries.append(row)
                elif len(row) == 3:
                    entries.append([row[0], row[1], row[2], ""])
                elif len(row) == 2:
                    entries.append([row[0], "unknown", row[1], ""])
    # Filter out blacklisted UUIDs
    return [row for row in entries if not is_uuid_blacklisted(row[1])]

# Load link name mapping once at startup
LINK_NAME_MAP = {}
try:
    with open('link_names.json', 'r', encoding='utf-8') as f:
        LINK_NAME_MAP = json.load(f)
except Exception:
    LINK_NAME_MAP = {}

def get_link_name(link):
    import re
    base_link = link.split('|')[0]
    # 1. Exact match (full URL)
    if base_link in LINK_NAME_MAP:
        return LINK_NAME_MAP[base_link]
    if link in LINK_NAME_MAP:
        return LINK_NAME_MAP[link]
    # 2. Substring/pattern match (e.g. "premium767" in link)
    for key, name in LINK_NAME_MAP.items():
        if key in base_link:
            # If the value contains {num}, try to extract the number after the key or before .m3u8
            if "{num}" in name:
                m = re.search(rf"{re.escape(key)}(\d+)", base_link)
                if m:
                    return name.format(num=m.group(1))
                m2 = re.search(r"/(\d+)\.m3u8", base_link)
                if m2:
                    return name.format(num=m2.group(1))
            # If the value contains {id}, extract the segment before /playlist.m3u8
            if "{id}" in name:
                m = re.search(r"/([^/]+)/playlist\.m3u8", base_link)
                if m:
                    return name.format(id=m.group(1))
            return name
    # 3. Generic: if link matches /something/playlist.m3u8, extract "something"
    m = re.search(r"/([^/]+)/playlist\.m3u8", base_link)
    if m:
        return m.group(1)
    return link  # fallback to raw link

@app.route('/dashboard')
def dashboard():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import defaultdict, Counter

    # Only include entries with a non-empty link for all stats except Recent Activity
    entries_with_link = [row for row in entries if row[3]]

    # Usage summary tables (only real link hits)
    addon_users = defaultdict(set)
    for addon, user_uuid, timestamp, link in entries_with_link:
        addon_users[addon].add(user_uuid)
    addon_hits = Counter([row[0] for row in entries_with_link])
    table_html = "".join(
        f"<tr><td>{addon}</td><td>{len(users)} ({', '.join(users)})</td><td>{addon_hits[addon]}</td></tr>"
        for addon, users in addon_users.items()
    )

    # Recent Activity (show only entries with a link)
    recent_entries = [row for row in entries_with_link][-50:][::-1]
    recent_html = "".join(
        f"<tr><td>{addon}</td><td>{user_uuid}</td><td>{timestamp}</td></tr>"
        for addon, user_uuid, timestamp, link in recent_entries
    )

    # Per-user stats with most-used addon (only real link hits)
    user_hits = Counter([row[1] for row in entries_with_link])
    top_users = user_hits.most_common(30)
    user_addon_hits = Counter((row[1], row[0]) for row in entries_with_link)
    top_users_html = ""
    for user, count in top_users:
        addons_for_user = [(addon, c) for (u, addon), c in user_addon_hits.items() if u == user]
        if addons_for_user:
            most_used_addon, addon_count = max(addons_for_user, key=lambda x: x[1])
        else:
            most_used_addon, addon_count = "-", 0
        top_users_html += f"<tr><td>{user}</td><td>{count}</td><td>{most_used_addon}</td></tr>"

    # Trending links with all addons and counts per link (only real link hits)
    link_addon_pairs = [(row[3], row[0]) for row in entries_with_link]
    link_counts = Counter([link for link, addon in link_addon_pairs])
    top_links = link_counts.most_common(30)
    top_links_html = ""
    for link, count in top_links:
        addons_for_link = [addon for l, addon in link_addon_pairs if l == link]
        addon_counts = Counter(addons_for_link)
        addon_counts_str = ", ".join(f"{addon} ({c})" for addon, c in addon_counts.most_common())
        display_link = get_link_name(link)
        top_links_html += f"<tr><td>{display_link}</td><td>{addon_counts_str}</td><td>{count}</td></tr>"

    import time
    cache_bust = str(int(time.time()))
    html = f"""
    <html>
    <head><title>Addon Usage Dashboard</title></head>
    <body>
        <h1>Addon Usage Summary</h1>
        <table border="1">
            <tr><th>Addon</th><th>Unique Users (UUIDs)</th><th>Total Hits</th></tr>
            {table_html}
        </table>
        <h2>Recent Activity (Last 50)</h2>
        <table border="1">
            <tr><th>Addon</th><th>User UUID</th><th>Timestamp</th></tr>
            {recent_html}
        </table>
        <h2>Most Active Users</h2>
        <table border="1">
            <tr><th>User UUID</th><th>Total Hits</th><th>Most Used Addon</th></tr>
            {top_users_html}
        </table>
        <h2>User Activity Over Time</h2>
        <img src="/user_activity_chart.png?cb={cache_bust}" alt="User Activity Over Time">
        <h2>Addon Usage Trends (Daily)</h2>
        <img src="/addon_trends_chart.png?cb={cache_bust}" alt="Addon Usage Trends">
        <h2>Unique User Chart</h2>
        <img src="/chart.png?cb={cache_bust}" alt="Addon Unique Users Chart">
        <h2>Total Hits Chart</h2>
        <img src="/hits_chart.png?cb={cache_bust}" alt="Addon Total Hits Chart">
        <h2>Usage Over Time (Daily)</h2>
        <img src="/daily_chart.png?cb={cache_bust}" alt="Daily Usage Chart">
        <h2>Top 5 Most Used Addons</h2>
        <img src="/topn_chart.png?cb={cache_bust}" alt="Top N Addons Chart">
        <h2>Pie Chart of Usage Distribution</h2>
        <img src="/pie_chart.png?cb={cache_bust}" alt="Pie Chart">
        <h2>Average Usage Per User Per Addon</h2>
        <img src="/avg_per_user_chart.png?cb={cache_bust}" alt="Average Usage Per User">
        <h2>Heatmap: Usage by Hour of Day</h2>
        <img src="/hour_heatmap.png?cb={cache_bust}" alt="Hour Heatmap">
        <h2>Trending Extractor Links</h2>
        <table border="1">
            <tr><th>Link</th><th>Addon</th><th>Hits</th></tr>
            {top_links_html}
        </table>
        <h2>Trending Links Chart</h2>
        <img src="/trending_links_chart.png?cb={cache_bust}" alt="Trending Links Chart">
        
        
    </body>
    </html>
    """
    return render_template_string(html)

@app.route('/chart.png')
def chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import defaultdict
    addon_users = defaultdict(set)
    for addon, user_uuid, timestamp, link in entries:
        addon_users[addon].add(user_uuid)
    addons = list(addon_users.keys())
    counts = [len(users) for users in addon_users.values()]
    plt.figure(figsize=(8,4))
    try:
        plt.bar([str(a) for a in addons], counts, color='skyblue')
        plt.xlabel('Addon')
        plt.ylabel('Unique Users')
        plt.title('Addon Unique Users')
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()  # Ensure renderer is available
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/hits_chart.png')
def hits_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import Counter
    addon_hits = Counter([row[0] for row in entries])
    addons = list(addon_hits.keys())
    counts = list(addon_hits.values())
    plt.figure(figsize=(8,4))
    try:
        plt.bar([str(a) for a in addons], counts, color='orange')
        plt.xlabel('Addon')
        plt.ylabel('Total Hits')
        plt.title('Addon Total Hits')
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in hits_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/daily_chart.png')
def daily_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import Counter
    dates = [row[2][:10] for row in entries]  # YYYY-MM-DD
    date_counts = Counter(dates)
    if not date_counts:
        return empty_chart_message()
    sorted_dates = sorted(date_counts)
    counts = [date_counts[d] for d in sorted_dates]
    plt.figure(figsize=(8,4))
    try:
        plt.plot(sorted_dates, counts, marker='o')
        plt.xlabel('Date')
        plt.ylabel('Hits')
        plt.title('Addon Usage Over Time (Daily)')
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in daily_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/topn_chart.png')
def topn_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import Counter
    addon_hits = Counter([row[0] for row in entries])
    topn = addon_hits.most_common(10)
    if not topn:
        return empty_chart_message()
    addons, counts = zip(*topn)
    plt.figure(figsize=(8,4))
    try:
        plt.bar([str(a) for a in addons], counts, color='green')
        plt.xlabel('Addon')
        plt.ylabel('Total Hits')
        plt.title('Top 5 Most Used Addons')
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in topn_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/pie_chart.png')
def pie_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import Counter
    addon_hits = Counter([row[0] for row in entries])
    if not addon_hits:
        return empty_chart_message()
    addons = list(addon_hits.keys())
    counts = list(addon_hits.values())
    plt.figure(figsize=(6,6))
    try:
        plt.pie(counts, labels=[str(a) for a in addons], autopct='%1.1f%%', startangle=140)
        plt.title('Addon Usage Distribution')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in pie_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/avg_per_user_chart.png')
def avg_per_user_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import defaultdict, Counter
    addon_users = defaultdict(set)
    for addon, user_uuid, timestamp, link in entries:
        addon_users[addon].add(user_uuid)
    addon_hits = Counter([row[0] for row in entries])
    addons = []
    avgs = []
    for addon in addon_users:
        user_count = len(addon_users[addon])
        hit_count = addon_hits[addon]
        avg = hit_count / user_count if user_count else 0
        addons.append(addon)
        avgs.append(avg)
    if not addons:
        return empty_chart_message()
    plt.figure(figsize=(8,4))
    try:
        plt.bar([str(a) for a in addons], avgs, color='purple')
        plt.xlabel('Addon')
        plt.ylabel('Avg Hits per User')
        plt.title('Average Usage Per User Per Addon')
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in avg_per_user_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/hour_heatmap.png')
def hour_heatmap_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    heatmap = np.zeros((7,24), dtype=int)
    for _, _, timestamp, _ in entries:
        try:
            dt = datetime.fromisoformat(timestamp)
            dow = dt.weekday()
            hour = dt.hour
            heatmap[dow, hour] += 1
        except Exception:
            continue
    plt.figure(figsize=(12,4))
    try:
        im = plt.imshow(heatmap, aspect='auto', cmap='YlOrRd')
        plt.colorbar(im, label='Hits')
        plt.xlabel('Hour of Day')
        plt.ylabel('Day of Week')
        plt.title('Usage Heatmap by Hour and Day')
        plt.yticks(ticks=range(7), labels=['Mon','Tue','Wed','Thu','Fri','Sat','Sun'])
        plt.xticks(ticks=range(0,24,2))
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in hour_heatmap_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/user_activity_chart.png')
def user_activity_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import Counter
    # Group by date
    dates = [row[2][:10] for row in entries]
    date_counts = Counter(dates)
    if not date_counts:
        return empty_chart_message()
    sorted_dates = sorted(date_counts)
    counts = [date_counts[d] for d in sorted_dates]
    plt.figure(figsize=(8,4))
    try:
        plt.plot(sorted_dates, counts, marker='o', color='brown')
        plt.xlabel('Date')
        plt.ylabel('Total User Hits')
        plt.title('User Activity Over Time')
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in user_activity_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/addon_trends_chart.png')
def addon_trends_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import defaultdict, Counter
    addon_date_counts = defaultdict(lambda: Counter())
    for addon, _, timestamp, _ in entries:
        date = timestamp[:10]
        addon_date_counts[addon][date] += 1
    if not addon_date_counts:
        return empty_chart_message()
    # Only plot top 3 addons for clarity
    top_addons = Counter([row[0] for row in entries]).most_common(5)
    top_addon_names = [a for a, _ in top_addons]
    all_dates = sorted({timestamp[:10] for _, _, timestamp, _ in entries})
    plt.figure(figsize=(10,5))
    try:
        for addon in top_addon_names:
            counts = [addon_date_counts[addon][d] for d in all_dates]
            plt.plot(all_dates, counts, marker='o', label=addon)
        plt.xlabel('Date')
        plt.ylabel('Hits')
        plt.title('Addon Usage Trends (Top 3 Addons)')
        plt.legend()
        plt.xticks(rotation=45, ha='right')
        plt.tight_layout()
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in addon_trends_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

@app.route('/trending_links_chart.png')
def trending_links_chart_png():
    log_file = '/srv/http/Requests/addon_log.csv'
    entries = parse_entries(log_file)
    from collections import Counter
    link_counts = Counter([row[3] for row in entries if row[3]])
    if not link_counts:
        return empty_chart_message()
    top_links = link_counts.most_common(50)
    links, counts = zip(*top_links)
    # Use display names for links
    display_links = [get_link_name(l) for l in links]
    plt.figure(figsize=(12, 4 + 0.5 * len(links)))  # Make chart taller for more links
    try:
        y_pos = range(len(display_links))
        plt.barh(y_pos, counts, color='teal')
        plt.yticks(y_pos, display_links, fontsize=9)
        plt.xlabel('Hits')
        plt.ylabel('Link')
        plt.title('Trending Extractor Links')
        plt.tight_layout()
        plt.gca().invert_yaxis()  # Highest at top
        plt.draw()
        img = io.BytesIO()
        plt.savefig(img, format='png', bbox_inches='tight')
        img.seek(0)
        return send_file(img, mimetype='image/png')
    except Exception as e:
        logging.error(f"Chart error in trending_links_chart_png: {str(e)}")
        return empty_chart_message(f"Chart error: {str(e)}")
    finally:
        plt.close()

if __name__ == '__main__':
    app.run(host='0.0.0.0', port=5001, debug=False, use_reloader=False, threaded=False)

